// ****************************************************************************
//  Project:        GUYMAGER
// ****************************************************************************
//  Programmer:     Guy Voncken
//                  Police Grand-Ducale
//                  Service de Police Judiciaire
//                  Section Nouvelles Technologies
// ****************************************************************************
//  Module:         The DeviceList represents the central data structure of
//                  the application. Its contents are displayed in the main
//                  window.
// ****************************************************************************

#include <float.h>

#include <QString>
#include <QList>

#include "toolconstants.h"

#include "common.h"
#include "config.h"
#include "mainwindow.h"

// ---------------------------
//         t_Device
// ---------------------------

static const int DEVICE_MIN_RUNNING_SECONDS = 20;  // The minmum number of acquisition seconds before any estimation can be made

void t_Device::Initialise (void)
{
   static bool Initialised = false;

   if (!Initialised)
   {
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DEVICE_SERNR_MATCH_MODEL_MISMATCH ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DEVICE_SERNR_MATCH_LENGTH_MISMATCH))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DEVICE_BAD_DEVICE_STATE           ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DEVICE_BAD_ACQUIRE_STATE          ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DEVICE_BAD_ABORTREASON            ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_DEVICE_NOT_CLEAN                  ))
      Initialised = true;
   }

//   this->SerialNumber      = QString();
//   this->LinuxDevice       = QString();
//   this->Model             = QString();
   this->Local             = false;
   this->SectorSize        = 0;
   this->SectorSizePhys    = 0;
   this->Size              = 0;
   this->Removable         = false;

   this->AcquireState      = Idle;
//   this->Acquisition.Message       = QString();
   this->DeviceState       = Connected;
   this->AbortRequest      = false;
   this->AbortReason       = None;
   this->DeleteAfterAbort  = false;
   this->pFifoRead         = NULL;
   this->pThreadRead       = NULL;
   this->pFileSrc          = NULL;
   this->CurrentReadPos    = 0;
//   this->BadSectors
   this->FallbackMode      = false;
   this->StartTimestamp    = QDateTime();

   this->pThreadWrite      = NULL;
   this->CurrentWritePos   = 0;
   this->pFifoWrite        = NULL;
   this->StopTimestamp     = QDateTime();

   this->pThreadMD5        = NULL;
   this->pFifoMD5In        = NULL;
   this->pFifoMD5Out       = NULL;
   memset (this->MD5Digest, 0, sizeof(this->MD5Digest));

   this->pFifoCompressIn   = NULL;
   this->pFifoCompressOut  = NULL;

//   this->Acquisition.Path          = QString();
//   this->Acquisition.ImageFilename = QString();
//   this->Acquisition.InfoFilename  = QString();
   this->Acquisition.Format        = t_File::NotSet;
   this->Acquisition.CalcMD5       = false;
   this->Acquisition.VerifyMD5     = false;
//   this->Acquisition.CaseNumber    = QString();
//   this->Acquisition.EvidenceNumber= QString();
//   this->Acquisition.Examiner      = QString();
//   this->Acquisition.Description   = QString();
//   this->Acquisition.Notes         = QString();

   this->Info.SetDevice(this);

   this->PrevTimestamp     = QTime();
   this->PrevSpeed         = 0.0;
   this->Checked           = false;
}

void t_Device::InitialiseDeviceSpecific (const QString &SerialNumber, const QString &LinuxDevice, const QString &Model,
                                         quint64 SectorSize, quint64 SectorSizePhys, quint64 Size)
{                                        //lint !e578: Declaration of symbol 'Xxx' hides symbol 't_Device::Xxx'
   this->SerialNumber   = SerialNumber;
   this->LinuxDevice    = LinuxDevice;
   this->Model          = Model;

   this->SectorSize     = SectorSize;
   this->SectorSizePhys = SectorSizePhys;
   this->Size           = Size;
//   this->Size           = 10737418240ULL;  // For faster testing (we make guymager believe that the device is smaller)
}

t_Device::t_Device (const QString &SerialNumber, const QString &LinuxDevice, const QString &Model,
                    quint64 SectorSize, quint64 SectorSizePhys, quint64 Size)
{                   //lint !e578: Declaration of symbol 'Xxx' hides symbol 't_Device::Xxx'
   Initialise ();
   InitialiseDeviceSpecific (SerialNumber, LinuxDevice, Model, SectorSize, SectorSizePhys, Size);
}

t_Device::t_Device (const QString &SerialNumber, const PedDevice *pPedDev)
{                   //lint !e578: Declaration of symbol 'Xxx' hides symbol 't_Device::Xxx'
   Initialise ();
   InitialiseDeviceSpecific (SerialNumber, pPedDev->path, pPedDev->model,
                                           (quint64) pPedDev->sector_size, (quint64) pPedDev->phys_sector_size,
                                           (quint64) pPedDev->length * (quint64) pPedDev->sector_size);
}

t_Device::~t_Device()
{
   if (pThreadRead   || pFifoRead   ||
       pThreadMD5    || pFifoMD5In  ||
       pThreadWrite  || pFifoMD5Out ||
       pFileSrc      || pFifoWrite )
      CHK_EXIT (ERROR_DEVICE_NOT_CLEAN)

   pThreadRead  = NULL; pFifoRead   = NULL;
   pThreadMD5   = NULL; pFifoMD5In  = NULL;
   pThreadWrite = NULL; pFifoMD5Out = NULL;
   pFileSrc     = NULL; pFifoWrite  = NULL;
}

bool t_Device::HasHashThread (void) const
{
   return Acquisition.CalcMD5 && CONFIG (UseSeparateThreadForMD5);
}

bool t_Device::HasCompressionThreads (void) const
{
   return (Acquisition.Format == t_File::EWF) && (CONFIG (CompressionThreads) > 0);
}


//lint -save -e818 pDevice could have declared as pointing to const

QVariant t_Device::GetSerialNumber (t_pDevice pDevice)
{
   return pDevice->SerialNumber;
}

QVariant t_Device::GetLinuxDevice (t_pDevice pDevice)
{
   return pDevice->LinuxDevice;
}

QVariant t_Device::GetModel (t_pDevice pDevice)
{
   return pDevice->Model;
}

QVariant t_Device::GetDeviceState (t_pDevice pDevice)
{
   switch (pDevice->DeviceState)
   {
      case Connected   : return t_MainWindow::tr("Connected"   );
      case Disconnected: return t_MainWindow::tr("Disconnected");
      default          : CHK_EXIT (ERROR_DEVICE_BAD_DEVICE_STATE)
   }
   return QVariant();
}

QVariant t_Device::GetAcquireState (t_pDevice pDevice)
{
   QString State;

   switch (pDevice->AcquireState)
   {
      case Idle    : if (pDevice->Local)
                          State = t_MainWindow::tr("Local device");
                     else State = t_MainWindow::tr("Idle");
                     break;
      case Running : State = t_MainWindow::tr("Running" ); break;
      case Cleanup : State = t_MainWindow::tr("Cleanup" ); break;
      case Paused  : State = t_MainWindow::tr("Paused"  ); break;
      case Finished: State = t_MainWindow::tr("Finished"); break;
      case Aborted : switch (pDevice->AbortReason)
                     {
                        case t_Device::None                : State = t_MainWindow::tr("Aborted - Error: Reason is 'none'" ); break;
                        case t_Device::UserRequest         : State = t_MainWindow::tr("Aborted by user" );                   break;
                        case t_Device::ThreadWriteFileError: State = t_MainWindow::tr("Aborted - Image file write error" );  break;
                        case t_Device::ThreadReadFileError : State = t_MainWindow::tr("Aborted - Device read error" );       break;
                        default                            : CHK_EXIT (ERROR_DEVICE_BAD_ABORTREASON)
                     }
                     break;
      default      : CHK_EXIT (ERROR_DEVICE_BAD_ACQUIRE_STATE)
   }
   if (pDevice->AcquireState != Aborted)
   {
      QString Msg;
      CHK_EXIT (pDevice->GetMessage (Msg))
      if (!Msg.isEmpty())
         State += " - " + Msg;
   }

   return State;
}

QVariant t_Device::GetSectorSize (t_pDevice pDevice)
{
   return pDevice->SectorSize;
}

QVariant t_Device::GetSectorSizePhys (t_pDevice pDevice)
{
   return pDevice->SectorSizePhys;
}

QVariant t_Device::GetSize (t_pDevice pDevice)
{
   return pDevice->Size;
}

QVariant t_Device::GetSizeHuman (t_pDevice pDevice)
{
   QString SizeHuman;
   QString Unit;
   QLocale Locale;
   double  Sz;

   Sz = pDevice->Size;
   if (Sz >= 1024) { Sz = Sz / 1024.0; Unit = "K"; }
   if (Sz >= 1024) { Sz = Sz / 1024.0; Unit = "M"; }
   if (Sz >= 1024) { Sz = Sz / 1024.0; Unit = "G"; }
   if (Sz >= 1024) { Sz = Sz / 1024.0; Unit = "T"; }
   if (Sz >= 1024) { Sz = Sz / 1024.0; Unit = "P"; }

   SizeHuman  = Locale.toString (Sz, 'f', 1);
   SizeHuman += Unit;

   return SizeHuman;
}

QVariant t_Device::GetBadSectorCount (t_pDevice pDevice)
{
   if (!pDevice->StartTimestamp.isNull())
        return pDevice->GetBadSectorCount ();
   else return "";
}

QVariant t_Device::GetProgress (t_pDevice pDevice)
{
   if (!pDevice->StartTimestamp.isNull())
        return (double)pDevice->GetCurrentWritePos() / (double)pDevice->Size;
   else return -DBL_MAX;
}

QVariant t_Device::GetCurrentSpeed (t_pDevice pDevice)
{
   QLocale    Locale;
   QString    Result;
   QTime      Now;
   double     MBs;
   double     Speed;
   int        MilliSeconds;
   quint64    WritePos;

   if (pDevice->AcquireState == Running)  // Don't display anything if no acquisition is running
   {
      Now = QTime::currentTime();
      WritePos = pDevice->GetCurrentWritePos();
      if (!pDevice->PrevTimestamp.isNull())
      {
         MilliSeconds = pDevice->PrevTimestamp.msecsTo (Now); // As this fn is called within 1s-interval, time_t would not be precise enough
         if (MilliSeconds > 0)
         {
            MBs   = (double) (WritePos - pDevice->PrevWritePos) / BYTES_PER_MEGABYTE;
            Speed = (1000.0*MBs) / MilliSeconds;
            pDevice->PrevSpeed = (pDevice->PrevSpeed + Speed) / 2.0;
         }
         Result = Locale.toString (pDevice->PrevSpeed, 'f', 2);
      }
      pDevice->PrevTimestamp = Now;
      pDevice->PrevWritePos  = WritePos;
   }

   return Result;
}

QVariant t_Device::GetAverageSpeed (t_pDevice pDevice)
{
   QLocale Locale;
   QString Result;
   double  MBs;
   int     Seconds;

   if (pDevice->StartTimestamp.isNull())
      return QString();

   if (pDevice->StopTimestamp.isNull()) // As long as the stop timestamp is Null, the acquisition is still running
        Seconds = pDevice->StartTimestamp.secsTo (QDateTime::currentDateTime());
   else Seconds = pDevice->StartTimestamp.secsTo (pDevice->StopTimestamp);

   if ((pDevice->StopTimestamp.isNull()) && (Seconds < DEVICE_MIN_RUNNING_SECONDS))
      return "--";

   MBs = ((double) pDevice->GetCurrentWritePos()) / BYTES_PER_MEGABYTE;
   Result = Locale.toString (MBs/Seconds, 'f', 2);

   return Result;
}

QVariant t_Device::GetRemaining (t_pDevice pDevice)
{
   QString Result;
   char    Buff[10];
   int     RunningSeconds;
   time_t  Now;
   int     hh, mm, ss;

   if (pDevice->AcquireState == Running)  // Don't display anything if no acquisition is running
   {
      time (&Now);
      RunningSeconds = pDevice->StartTimestamp.secsTo (QDateTime::currentDateTime());
      if (RunningSeconds < DEVICE_MIN_RUNNING_SECONDS)
      {
         Result = "--";
      }
      else
      {
         ss  = (int) (((double)pDevice->Size / (double)pDevice->GetCurrentWritePos()) * RunningSeconds); // Estimated total time
         ss -= RunningSeconds;                                                                           // Estimated remaining time
         hh = ss / SECONDS_PER_HOUR;   ss %= SECONDS_PER_HOUR;
         mm = ss / SECONDS_PER_MINUTE; ss %= SECONDS_PER_MINUTE;
         snprintf (Buff, sizeof(Buff), "%02d:%02d:%02d", hh, mm, ss);
         Result = Buff;
      }
   }

   return Result;
}

QVariant t_Device::GetFifoStatus (t_pDevice pDevice)
{
   QString Result;
   int     Count1, Count2, Count3;

   if (pDevice->AcquireState == Running)  // Don't display anything if no acquisition is running
   {
      if (!pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
      {
         CHK_EXIT (pDevice->pFifoRead->Count(Count1))
         Result = QString ("r %1 w") .arg (Count1);
      }
      else if (!pDevice->HasHashThread() && pDevice->HasCompressionThreads())
      {
         CHK_EXIT (pDevice->pFifoRead       ->Count(Count1))
         CHK_EXIT (pDevice->pFifoCompressOut->Count(Count2))
         Result = QString ("r %1 c %2 w") .arg (Count1) .arg(Count2);
      }
      else if (pDevice->HasHashThread() && !pDevice->HasCompressionThreads())
      {
         CHK_EXIT (pDevice->pFifoRead  ->Count(Count1))
         CHK_EXIT (pDevice->pFifoMD5Out->Count(Count2))
         Result = QString ("r %1 m %2 w") .arg (Count1) .arg(Count2);
      }
      else if (pDevice->HasHashThread() && pDevice->HasCompressionThreads())
      {
         CHK_EXIT (pDevice->pFifoRead       ->Count(Count1))
         CHK_EXIT (pDevice->pFifoMD5Out     ->Count(Count2))
         CHK_EXIT (pDevice->pFifoCompressOut->Count(Count3))
         Result = QString ("r %1 m %2 c %3 w") .arg (Count1) .arg(Count2) .arg(Count3);
      }
   }

   return Result;
}


//lint -restore


// ---------------------------
//        t_DeviceList
// ---------------------------

t_DeviceList::t_DeviceList(void)
   :QList<t_pDevice>()
{
   static bool Initialised = false;

   if (!Initialised)
   {
      Initialised = true;
      qRegisterMetaType<t_pDeviceList>("t_pDeviceList");
   }
}

t_DeviceList::~t_DeviceList()
{
   int i;

   for (i=0; i<count(); i++)
      delete at(i);
}

t_pDevice t_DeviceList::AppendNew (const QString &SerialNumber, const PedDevice *pPedDev)
{
   t_pDevice pDev;

   pDev = new t_Device (SerialNumber, pPedDev);
   append (pDev);

   return pDev;
}

t_pDevice t_DeviceList::AppendNew (const QString &SerialNumber, const QString &LinuxDevice, const QString &Model,
                                   quint64 SectorSize, quint64 SectorSizePhys, quint64 Size)
{
   t_pDevice pDev;

   pDev = new t_Device (SerialNumber, LinuxDevice, Model, SectorSize, SectorSizePhys, Size);
   append (pDev);

   return pDev;
}

//t_pDevice t_DeviceList::SearchLinuxDevice  (const QString &LinuxDevice)
//{
//   t_pDevice pDev;
//   int        i;
//
//   for (i=0; i<count(); i++)
//   {
//      pDev = at(i);
//      if (pDev->LinuxDevice == LinuxDevice)
//         return pDev;
//   }
//   return NULL;
//}
//
//t_pDevice t_DeviceList::SearchSerialNumber (const QString &SerialNumber)
//{
//   t_pDevice pDev;
//   int        i;
//
//   for (i=0; i<count(); i++)
//   {
//      pDev = at(i);
//      if (pDev->SerialNumber == SerialNumber)
//         return pDev;
//   }
//   return NULL;
//}

APIRET t_DeviceList::MatchDevice (t_pcDevice pDevCmp, t_pDevice &pDevMatch)
{
   bool Found = false;
   int  i;

   for (i=0; (i<count()) && !Found; i++)
   {
      pDevMatch = at(i);
      if (pDevMatch->SerialNumber.isEmpty())
      {
         Found = (pDevCmp->SerialNumber.isEmpty() && (pDevMatch->DeviceState == t_Device::Connected) &&
                                                     (pDevMatch->Model       == pDevCmp->Model     ) &&
                                                     (pDevMatch->Size        == pDevCmp->Size      ));
      }
      else
      {
         Found = (pDevMatch->SerialNumber == pDevCmp->SerialNumber);
         if (Found)
         {                                                                                            // If the serial numbers match, the whole
            if (pDevMatch->Model != pDevCmp->Model) return ERROR_DEVICE_SERNR_MATCH_MODEL_MISMATCH;   // rest must match as well. Otherwise,
            if (pDevMatch->Size  != pDevCmp->Size ) return ERROR_DEVICE_SERNR_MATCH_LENGTH_MISMATCH;  // something very strange is going on...
         }
      }
   }

   if (!Found)
      pDevMatch = NULL;

   return NO_ERROR;
}


