// ****************************************************************************
//  Project:        GUYMAGER
// ****************************************************************************
//  Programmer:     Guy Voncken
//                  Police Grand-Ducale
//                  Service de Police Judiciaire
//                  Section Nouvelles Technologies
// ****************************************************************************
//  Module:         Thread for reading data.
// ****************************************************************************

#include <errno.h>

#include <QtCore>

#include "common.h"
#include "config.h"
#include "device.h"
#include "threadread.h"
#include "threadwrite.h"

#include "fcntl.h"
//##define _GNU_SOURCE

#if (LIBEWF_VERSION < 20071111)
   #include "libewf.h"
#endif

const int THREADREAD_CHECK_CONNECTED_SLEEP = 1000;
const int THREADREAD_SLOWDOWN_SLEEP        =  700;


class t_ThreadReadLocal
{
   public:
      t_pDevice     pDevice;
      bool         *pSlowDownRequest;
      unsigned int   BlockSize;
};

t_ThreadRead::t_ThreadRead(void)
{
   CHK_EXIT (ERROR_THREADREAD_CONSTRUCTOR_NOT_SUPPORTED)
} //lint !e1401 not initialised

t_ThreadRead::t_ThreadRead (t_pDevice pDevice, bool *pSlowDownRequest)
{
   static bool Initialised = false;

   if (!Initialised)
   {
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_FILESRC_ALREADY_OPEN     ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_NO_DATA                  ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_DEVICE_DISCONNECTED      ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_UNEXPECTED_FAILURE       ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_CONSTRUCTOR_NOT_SUPPORTED))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_BLOCKSIZE                ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_DEVICE_ABORTED           ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_LIBEWF_FAILED            ))
      CHK_EXIT (TOOL_ERROR_REGISTER_CODE (ERROR_THREADREAD_BAD_FILE_HANDLE          ))
      Initialised = true;
   }

   pOwn = new t_ThreadReadLocal;
   pOwn->pDevice          = pDevice;
   pOwn->pSlowDownRequest = pSlowDownRequest;

   if (pDevice->HasCompressionThreads())
        pOwn->BlockSize = MULTITHREADED_COMPRESSION_CHUNK_SIZE;
   else pOwn->BlockSize = (unsigned int) CONFIG(FifoBlockSize);

   if (pOwn->BlockSize != (unsigned int) CONFIG(FifoBlockSize))
      LOG_INFO ("Configuration parameter FifoBlockSize has been set to %d (no other size supported when running with multithread EWF compression)", pOwn->BlockSize)

   CHK_QT_EXIT (connect (this, SIGNAL(finished()), this, SLOT(SlotFinished())))
}

t_ThreadRead::~t_ThreadRead (void)
{
   delete pOwn;
}

inline bool ThreadReadEOF (t_pcDevice pDevice)
{
//   LOG_INFO ("EOF check %Ld %Ld", pDevice->CurrentPos, pDevice->Size)
   return (pDevice->CurrentReadPos >= pDevice->Size);
}

static APIRET ThreadReadCheckDeviceExists (t_pcDevice pDevice, bool &DeviceExists)
{
   FILE *pFileTest;

   pFileTest = fopen64 (QSTR_TO_PSZ(pDevice->LinuxDevice), "r");
   DeviceExists = (pFileTest != NULL);
   if (DeviceExists)
      (void) fclose (pFileTest);

   return NO_ERROR;
}

static APIRET ThreadReadDeviceDisconnected (t_pDevice pDevice)
{
   LOG_INFO ("Switching %s to disconnected", QSTR_TO_PSZ(pDevice->LinuxDevice))
   if (pDevice->pFileSrc)
   {
      (void) fclose (pDevice->pFileSrc);
      pDevice->pFileSrc = NULL;
   }
   pDevice->DeviceState  = t_Device::Disconnected;
   pDevice->AcquireState = t_Device::Paused;

   return ERROR_THREADREAD_DEVICE_DISCONNECTED;
}

static APIRET ThreadReadAdjustSeek (t_pDevice pDevice)        // According to the GNU Clib docu, a seek should be done after read errors
{
   int rc;

   if (pDevice->pFileSrc == NULL)
      CHK (ERROR_THREADREAD_FILESRC_NOT_OPEN)

   if (ThreadReadEOF (pDevice))
      return NO_ERROR;

   rc = fseeko64 (pDevice->pFileSrc, (off64_t)pDevice->CurrentReadPos, SEEK_SET);
   if (rc)
      CHK_RET (ThreadReadDeviceDisconnected (pDevice))

   return NO_ERROR;
}

//static APIRET ThreadReadDirectMode (FILE *pFileSrc, bool Enable)
//{
//   int OldFlags, NewFlags, rc;
//   int FileDescriptor;
//
//   FileDescriptor = fileno (pFileSrc);
//   if (FileDescriptor == -1)
//      CHK (ERROR_THREADREAD_BAD_FILE_HANDLE)
//
//   OldFlags = fcntl (FileDescriptor, F_GETFL);
//   if (OldFlags > 0)
//   {
//      NewFlags = (Enable ? (OldFlags |  O_DIRECT)
//                         : (OldFlags & ~O_DIRECT));
//      if (NewFlags != OldFlags)
//      {
//         rc = fcntl (FileDescriptor, F_SETFL, NewFlags);
//         if (rc == -1)
//            LOG_INFO ("Setting flags with fcntl failed. Direct mode state unknown.")
//      }
//      LOG_INFO ("Direct mode %s", Enable ? "enabled" : "disabled")
//   }
//   else
//   {
//      LOG_INFO ("Reading flags with fcntl failed. Direct mode remains unchanged")
//   }
//
//   return NO_ERROR;
//}

static APIRET ThreadReadBlock0 (t_pDevice pDevice, unsigned char *pBuffer, unsigned int Size)
{
   bool DeviceExists;
   int  Read;

   errno = 0;
   if (pDevice->pFileSrc == NULL)
   {
      pDevice->pFileSrc = fopen64 (QSTR_TO_PSZ(pDevice->LinuxDevice), "r");
      if (!pDevice->pFileSrc)
         CHK_RET (ThreadReadDeviceDisconnected (pDevice))
      CHK_RET (ThreadReadAdjustSeek (pDevice))

//      CHK (ThreadReadDirectMode (pDevice->pFileSrc, true))
   }

   Read = (int) fread (pBuffer, Size, 1, pDevice->pFileSrc); //lint !e732 !e712 loss of sign/precision arg 2

// Bad sector simulation
//   if ((pDevice->CurrentReadPos > 500000) && (pDevice->CurrentReadPos < 600000) ||
//       (pDevice->CurrentReadPos > 700000) && (pDevice->CurrentReadPos < 740000) ||
//       (pDevice->CurrentReadPos == 0) || (pDevice->CurrentReadPos == 512) || (pDevice->CurrentReadPos == 1024))
//      Read = 0;

   switch (Read)
   {
      case  1: pDevice->CurrentReadPos += Size;
               return NO_ERROR;

      case  0:
//               LOG_INFO ("Read error at position %lld", pDevice->CurrentReadPos)

               CHK (ThreadReadCheckDeviceExists (pDevice, DeviceExists))
               if (DeviceExists)
               {
//                  LOG_DEBUG ("Device still exists")
                  CHK_RET (ThreadReadAdjustSeek (pDevice))
//                  LOG_DEBUG ("Seek adjusted")
                  return ERROR_THREADREAD_NO_DATA;
               }
               else
               {
                  CHK_RET (ThreadReadDeviceDisconnected (pDevice))
               }
               break;

      default: CHK_CONST (ERROR_THREADREAD_UNEXPECTED_FAILURE)
   }

   return NO_ERROR;
}


APIRET t_ThreadRead::ThreadReadBlock (t_pFifoBlock &pFifoBlock)
{
   t_pDevice      pDevice;
   unsigned char *pBuffer;
   quint64         BytesToRead;
   quint64         BytesRead = 0;
   quint64         RemainingSize;
   quint64         Sector;
   quint64         PrevReadPos;
   unsigned int    ReadTry;
   APIRET          RcRead;
   APIRET          RcSeek;

   pDevice       = pOwn->pDevice;
   PrevReadPos   = pDevice->CurrentReadPos;
   RemainingSize = pDevice->Size - pDevice->CurrentReadPos;
   if ((quint64)pOwn->BlockSize > RemainingSize)
        BytesToRead = RemainingSize;
   else BytesToRead = pOwn->BlockSize;

   if (BytesToRead > INT_MAX)
      CHK (ERROR_THREADREAD_BLOCKSIZE)

   if (pDevice->HasCompressionThreads())
        CHK (t_Fifo::CreateCompressionOptimised (pFifoBlock))
   else CHK (t_Fifo::Create (pFifoBlock, (unsigned int) BytesToRead))

   pFifoBlock->DataSize = (unsigned int) BytesToRead;
   pBuffer = pFifoBlock->Buffer;

   if (pDevice->FallbackMode)
        ReadTry = pDevice->SectorSize;
   else ReadTry = BytesToRead;
   do
   {
      RcRead = ThreadReadBlock0 (pDevice, pBuffer, ReadTry);
      switch (RcRead)
      {
         case NO_ERROR:
            BytesRead += ReadTry;
            pBuffer   += ReadTry;
            break;

         case ERROR_THREADREAD_NO_DATA:
            if (!pDevice->FallbackMode)
            {
               pDevice->FallbackMode = true;
//               ReadTry = (int) std::min (pDevice->SectorSize, pDevice->SectorSizePhys);
               ReadTry = pDevice->SectorSize;
//               CHK (ThreadReadDirectMode (pDevice->pFileSrc, true))
               LOG_INFO ("Device read error, switching %s to slow fallback mode. Reading sectors individually from now on.", QSTR_TO_PSZ(pDevice->LinuxDevice))
            }
            else
            {
               Sector = pDevice->CurrentReadPos / pDevice->SectorSize;
               pDevice->CurrentReadPos += ReadTry;
               RcSeek = ThreadReadAdjustSeek (pDevice); // If the seek still succeeds, then there's probably just a bad sector on
               switch (RcSeek)                          // the source device, let's replace it by a zero sector.
               {
                  case NO_ERROR:
                     memset (pBuffer, 0, ReadTry);
                     BytesRead += ReadTry;
                     pBuffer   += ReadTry;
                     LOG_INFO ("Read error on %s, sector %Ld, bad data replaced by zero block", QSTR_TO_PSZ(pDevice->LinuxDevice), Sector)
                     pDevice->AddBadSector (Sector);
                     break;
                  case ERROR_THREADREAD_DEVICE_DISCONNECTED:
                     break;
                  default:
                     CHK (RcSeek)
               }
            }
            break;

         case ERROR_THREADREAD_DEVICE_DISCONNECTED:
            break;

         default:
            CHK (RcRead)
      }
   } while ((BytesRead < BytesToRead) && (pDevice->DeviceState != t_Device::Disconnected) && !pDevice->AbortRequest);

   if (pDevice->DeviceState == t_Device::Disconnected)
   {
      CHK (t_Fifo::Destroy (pFifoBlock))
      pDevice->CurrentReadPos = PrevReadPos;

      return ERROR_THREADREAD_DEVICE_DISCONNECTED;
   }
   else if (pDevice->AbortRequest)
   {
      CHK (t_Fifo::Destroy (pFifoBlock))

      return ERROR_THREADREAD_DEVICE_ABORTED;
   }
   else if ((pDevice->FallbackMode) && (RcRead == NO_ERROR))
   {
      pDevice->FallbackMode = false;  // Switch to fast reading mode if last read was good
      LOG_INFO ("Device read ok again. Switching %s to fast block read.", QSTR_TO_PSZ(pDevice->LinuxDevice))
//      CHK (ThreadReadDirectMode (pDevice->pFileSrc, false))
   }

   return NO_ERROR;
}

void t_ThreadRead::run (void)
{
   t_pDevice      pDevice;
   t_pFifoBlock   pFifoBlock;
   APIRET          rc;
   int             rcf;
   bool            CalcMD5;
   MD5_CTX         MD5Context;
   quint64         Blocks = 0;
   #if (LIBEWF_VERSION < 20071111)
      LIBEWF_HANDLE *pEwfHandle= NULL;
   #endif

   LOG_INFO ("Acquisition of %s: Reading thread started", QSTR_TO_PSZ (pOwn->pDevice->LinuxDevice))

   pDevice = pOwn->pDevice;
   pDevice->pFileSrc       = NULL;
   pDevice->CurrentReadPos = 0;
   pDevice->FallbackMode   = false;
   CalcMD5 = pDevice->Acquisition.CalcMD5 && !CONFIG (UseSeparateThreadForMD5);

   if (CalcMD5)
   {
      (void) MD5_Init (&MD5Context);
      #if (LIBEWF_VERSION < 20071111)
         if (pDevice->HasCompressionThreads())
            CHK_EXIT (pDevice->pThreadWrite->GetpFileHandle ((void **)&pEwfHandle))
      #endif
   }

   do
   {
      pDevice->AcquireState = t_Device::Running;
      if (*(pOwn->pSlowDownRequest))
         msleep (THREADREAD_SLOWDOWN_SLEEP);

      rc = ThreadReadBlock (pFifoBlock);
      switch (rc)
      {
         case NO_ERROR:
            pFifoBlock->Nr = Blocks;

            if (CalcMD5)
            {
               (void) MD5_Update (&MD5Context, pFifoBlock->Buffer, (unsigned int)pFifoBlock->DataSize);
               #if (LIBEWF_VERSION < 20071111)
                  if (pDevice->HasCompressionThreads())
                  {
                     rc = libewf_raw_update_md5 (pEwfHandle, pFifoBlock->Buffer, (unsigned long) pFifoBlock->DataSize);
                     if (rc != 1)
                        CHK_EXIT (ERROR_THREADREAD_LIBEWF_FAILED)
                  }
               #endif
            }

            CHK_EXIT (pDevice->pFifoRead->Insert (pFifoBlock))
            Blocks++;
            break;

         case ERROR_THREADREAD_DEVICE_DISCONNECTED:
            while ((pDevice->DeviceState == t_Device::Disconnected) && !pDevice->AbortRequest)
               msleep (THREADREAD_CHECK_CONNECTED_SLEEP);
            if (pDevice->DeviceState == t_Device::Connected)
               pDevice->AcquireState = t_Device::Running;
            break;

         case ERROR_THREADREAD_DEVICE_ABORTED:
            break;
         default:
            CHK_EXIT (rc)
      }
   } while (!ThreadReadEOF(pDevice) && !pDevice->AbortRequest);

   if (ThreadReadEOF(pDevice))
      LOG_INFO ("All data has been read (device %s)", QSTR_TO_PSZ (pDevice->LinuxDevice))

   if (CalcMD5)
      (void) MD5_Final (pDevice->MD5Digest, &MD5Context);

   if (pDevice->pFileSrc)
   {
      rcf = fclose (pDevice->pFileSrc);
      pDevice->pFileSrc = NULL;
      if (rcf)
      {
         LOG_INFO ("Unexpected error on fclose, errno = %d", errno)
         pDevice->AbortReason  = t_Device::ThreadReadFileError;
         pDevice->AbortRequest = true;
      }
   }

   LOG_INFO ("Reading thread exits now (device %s, %Ld blocks processed, %Ld bytes read from device)", QSTR_TO_PSZ (pDevice->LinuxDevice), Blocks, pDevice->CurrentReadPos)
}

void t_ThreadRead::SlotFinished (void)
{
   emit SignalEnded (pOwn->pDevice);
}

